# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import physicsnemo
import torch
import torch.nn.functional as F
from physicsnemo.models.afno import AFNO


class PeriodicPad2d(torch.nn.Module):
    """
    pad longitudinal (left-right) circular
    and pad latitude (top-bottom) with zeros
    """

    def __init__(self, pad_width: int):
        super().__init__()
        self.pad_width = pad_width

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # pad left and right circular
        out = F.pad(x, (self.pad_width, self.pad_width, 0, 0), mode="circular")
        # pad top and bottom zeros
        out = F.pad(
            out, (0, 0, self.pad_width, self.pad_width), mode="constant", value=0
        )
        return out


class PrecipNet(physicsnemo.Module):
    def __init__(
        self,
        inp_shape: tuple,
        patch_size: tuple,
        in_channels: int,
        out_channels: int,
        embed_dim: int,
        depth: int,
        num_blocks: int,
        *args: tuple,
        **kwargs: dict,
    ):
        super().__init__()
        backbone = AFNO(
            inp_shape,
            in_channels,
            out_channels,
            patch_size,
            embed_dim,
            depth,
            num_blocks,
        )

        self.backbone = backbone
        self.ppad = PeriodicPad2d(1)
        self.conv = torch.nn.Conv2d(
            out_channels, out_channels, kernel_size=3, stride=1, padding=0, bias=True
        )
        self.act = torch.nn.ReLU()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.backbone(x)
        x = self.ppad(x)
        x = self.conv(x)
        x = self.act(x)
        return x



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from datetime import datetime
from typing import Any

import xarray as xr
from numpy import ndarray
from pandas import to_datetime

from earth2studio.utils.type import TimeArray, VariableArray


class DataArrayFile:
    """A local xarray dataarray file data source. This file should be compatable with
    xarray. For example, a netCDF file.

    Parameters
    ----------
    file_path : str
        Path to xarray data array compatible file.
    """

    def __init__(self, file_path: str, **xr_args: Any):
        self.file_path = file_path
        self.da = xr.open_dataarray(self.file_path, **xr_args)
        # self.da = xr.open_dataarray(self.file_path, **xr_args)

    def __call__(
        self,
        time: datetime | list[datetime] | TimeArray,
        variable: str | list[str] | VariableArray,
    ) -> xr.DataArray:
        """Function to get data.

        Parameters
        ----------
        time : datetime | list[datetime] | TimeArray
            Timestamps to return data for.
        variable : str | list[str] | VariableArray
            Strings or list of strings that refer to variables to return.

        Returns
        -------
        xr.DataArray
            Loaded data array
        """
        return self.da.sel(time=time, variable=variable)


class DataSetFile:
    """A local xarray dataset file data source. This file should be compatable with
    xarray. For example, a netCDF file.

    Parameters
    ----------
    file_path : str
        Path to xarray dataset compatible file.
    array_name : str
        Data array name in xarray dataset
    """

    def __init__(self, file_path: str, array_name: str, **xr_args: Any):
        self.file_path = file_path
        self.da = xr.open_dataset(self.file_path, **xr_args)[array_name]

    def __call__(
        self,
        time: datetime | list[datetime] | TimeArray,
        variable: str | list[str] | VariableArray,
    ) -> xr.DataArray:
        """Function to get data.

        Parameters
        ----------
        time : datetime | list[datetime] | TimeArray
            Timestamps to return data for.
        variable : str | list[str] | VariableArray
            Strings or list of strings that refer to variables to return.

        Returns
        -------
        xr.DataArray
            Loaded data array
        """
        return self.da.sel(time=time, variable=variable)


class DataArrayDirectory:
    """A local xarray dataarray directory data source. This file should be compatable with
    xarray. For example, a netCDF file. the structure of the directory should be like
    path/to/monthly/files
    |___2020
    |   |___2020_01.nc
    |   |___2020_02.nc
    |   |___ ...
    |
    |___2021
        |___2021_01.nc
        |___...

    Parameters
    ----------
    file_path : str
        Path to xarray data array compatible file.
    xr_args : Any
        Keyword arguments to send to the xarray opening method.
    """

    def __init__(self, dir_path: str, **xr_args: Any):
        self.dir_path = dir_path
        self.das: dict[str, dict[str, xr.DataArray]] = {}
        for yr in os.listdir(self.dir_path):
            yr_dir = os.path.join(self.dir_path, yr)
            if os.path.isdir(yr_dir):
                self.das[yr] = {}
                for fl in os.listdir(yr_dir):
                    pth = os.path.join(yr_dir, fl)
                    if os.path.isfile(pth):
                        try:
                            arr = xr.open_dataarray(pth, **xr_args)
                        except:  # noqa
                            continue
                        mon = fl.split(".")[0].split("_")[-1]
                        self.das[yr][mon] = arr

    def __call__(
        self,
        time: datetime | list[datetime] | TimeArray,
        variable: str | list[str] | VariableArray,
    ) -> xr.DataArray:
        """Function to get data.

        Parameters
        ----------
        time : datetime | list[datetime] | TimeArray
            Timestamps to return data for.
        variable : str | list[str] | VariableArray
            Strings or list of strings that refer to variables to return.

        Returns
        -------
        xr.DataArray
            Loaded data array
        """
        if not (isinstance(time, list) or isinstance(time, ndarray)):
            time = [time]
        if not (isinstance(variable, list) or isinstance(variable, ndarray)):
            variable = [variable]

        arrs = []
        for tt in time:
            yr = str(to_datetime(tt).year)
            mon = str(to_datetime(tt).month).zfill(2)
            arrs.append(self.das[yr][mon].sel(time=tt, variable=variable))

        return xr.concat(arrs, dim="time")



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime, timedelta
from typing import Protocol, runtime_checkable

import xarray as xr

from earth2studio.utils.type import LeadTimeArray, TimeArray, VariableArray


@runtime_checkable
class DataSource(Protocol):
    """Data source interface."""

    def __call__(
        self,
        time: datetime | list[datetime] | TimeArray,
        variable: str | list[str] | VariableArray,
    ) -> xr.DataArray:
        """Function to get data.

        Parameters
        ----------
        time : datetime | list[datetime] | TimeArray
            Datetime, list of datetimes or array of np.datetime64 to return data for.
        variable : str | list[str] | VariableArray
            String, list of strings or array of strings that refer to variables to
            return.

        Returns
        -------
        xr.DataArray
            An xarray data-array with the dimensions [time, variable, ....]. The coords
            should be provided. Time coordinate should be a datetime array and the
            variable coordinate should be array of strings with Earth2Studio variable
            ids.
        """
        pass

    async def fetch(  # type: ignore[override]
        self,
        time: datetime | list[datetime] | TimeArray,
        variable: str | list[str] | VariableArray,
    ) -> xr.DataArray:
        """Async function to get data. Async data sources support this.

        Parameters
        ----------
        time : datetime | list[datetime] | TimeArray
            Datetime, list of datetimes or array of np.datetime64 to return data for.
        variable : str | list[str] | VariableArray
            String, list of strings or array of strings that refer to variables to
            return.

        Returns
        -------
        xr.DataArray
            An xarray data-array with the dimensions [time, variable, ....]. The coords
            should be provided. Time coordinate should be a datetime array and the
            variable coordinate should be array of strings with Earth2Studio variable
            ids.
        """
        pass


@runtime_checkable
class ForecastSource(Protocol):
    """Forecast source interface"""

    def __call__(  # type: ignore[override]
        self,
        time: datetime | list[datetime] | TimeArray,
        lead_time: timedelta | list[timedelta] | LeadTimeArray,
        variable: str | list[str] | VariableArray,
    ) -> xr.DataArray:
        """Function to get data.

        Parameters
        ----------
        time : datetime | list[datetime] | TimeArray
            Datetime, list of datetimes or array of np.datetime64 to return data for.
        lead_time: timedelta | list[timedelta], LeadTimeArray
            Timedelta, list of timedeltas or array of np.timedelta that refers to the
            forecast lead time to fetch.
        variable : str | list[str] | VariableArray
            String, list of strings or array of strings that refer to variables to
            return.

        Returns
        -------
        xr.DataArray
            An xarray data-array with the dimensions [time, variable, lead_time, ...].
            The coords should be provided. Time coordinate should be a TimeArray,
            lead time coordinate a LeadTimeArray and the variable coordinate should be
            an array of strings with Earth2Studio variable ids.
        """
        pass

    async def fetch(  # type: ignore[override]
        self,
        time: datetime | list[datetime] | TimeArray,
        lead_time: timedelta | list[timedelta] | LeadTimeArray,
        variable: str | list[str] | VariableArray,
    ) -> xr.DataArray:
        """Async function to get data. Async forecast sources support this.

        Parameters
        ----------
        time : datetime | list[datetime] | TimeArray
            Datetime, list of datetimes or array of np.datetime64 to return data for.
        lead_time: timedelta | list[timedelta], LeadTimeArray
            Timedelta, list of timedeltas or array of np.timedelta that refers to the
            forecast lead time to fetch.
        variable : str | list[str] | VariableArray
            String, list of strings or array of strings that refer to variables to
            return.

        Returns
        -------
        xr.DataArray
            An xarray data-array with the dimensions [time, variable, lead_time, ...].
            The coords should be provided. Time coordinate should be a TimeArray,
            lead time coordinate a LeadTimeArray and the variable coordinate should be
            an array of strings with Earth2Studio variable ids.
        """
        pass



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from collections.abc import Generator, Iterator
from datetime import datetime

import numpy as np
import torch
import xarray as xr

try:
    from physicsnemo import Module as PhysicsNemoModule
    from physicsnemo.utils.zenith_angle import cos_zenith_angle
except ImportError:
    PhysicsNemoModule = None
    cos_zenith_angle = None

from earth2studio.models.auto import AutoModelMixin, Package
from earth2studio.models.batch import batch_coords, batch_func
from earth2studio.models.px.base import PrognosticModel
from earth2studio.models.px.utils import PrognosticMixin
from earth2studio.utils import check_extra_imports
from earth2studio.utils.coords import CoordSystem, map_coords

VARIABLES = [
    "u10m",
    "v10m",
    "u100m",
    "v100m",
    "t2m",
    "sp",
    "msl",
    "tcwv",
    "u50",
    "u100",
    "u150",
    "u200",
    "u250",
    "u300",
    "u400",
    "u500",
    "u600",
    "u700",
    "u850",
    "u925",
    "u1000",
    "v50",
    "v100",
    "v150",
    "v200",
    "v250",
    "v300",
    "v400",
    "v500",
    "v600",
    "v700",
    "v850",
    "v925",
    "v1000",
    "z50",
    "z100",
    "z150",
    "z200",
    "z250",
    "z300",
    "z400",
    "z500",
    "z600",
    "z700",
    "z850",
    "z925",
    "z1000",
    "t50",
    "t100",
    "t150",
    "t200",
    "t250",
    "t300",
    "t400",
    "t500",
    "t600",
    "t700",
    "t850",
    "t925",
    "t1000",
    "q50",
    "q100",
    "q150",
    "q200",
    "q250",
    "q300",
    "q400",
    "q500",
    "q600",
    "q700",
    "q850",
    "q925",
    "q1000",
]


@check_extra_imports("interp-modafno", [PhysicsNemoModule, cos_zenith_angle])
class InterpModAFNO(torch.nn.Module, AutoModelMixin, PrognosticMixin):
    """ModAFNO interpolation for global prognostic models. Interpolates a forecast model
    to a shorter time-step size (by default from 6 to 1 hour). Operates on 0.25 degree
    lat-lon equirectangular grid with 73 variables.

    Note
    ----
    For more information on the model, please refer to:

    - https://catalog.ngc.nvidia.com/orgs/nvidia/teams/earth-2/models/afno_dx_fi-v1-era5
    - https://arxiv.org/abs/2410.18904

    Warning
    -------
    The model requires a base forecast model to be set before execution. This can be
    done by setting the `px_model` attribute or using the `load_model` method.

    Parameters
    ----------
    interp_model : torch.nn.Module
        The interpolation model that performs the time interpolation
    center : torch.Tensor
        Model center normalization tensor
    scale : torch.Tensor
        Model scale normalization tensors
    geop : torch.Tensor
        Geopotential height data used as a static feature
    lsm : torch.Tensor
        Land-sea mask data used as a static feature
    px_model : PrognosticModel, optional
        The base forecast model that produces the coarse time resolution forecasts. If
        not provide, should be set by the user before executing the model, by default
        None.
    num_interp_steps : int, optional
        Number of interpolation steps to perform between forecast steps, by default 6
    """

    def __init__(
        self,
        interp_model: torch.nn.Module,
        center: torch.Tensor,
        scale: torch.Tensor,
        geop: torch.Tensor,
        lsm: torch.Tensor,
        px_model: PrognosticModel | None = None,
        num_interp_steps: int = 6,
    ) -> None:
        super().__init__()
        self.px_model = px_model
        self.interp_model = interp_model
        self.num_interp_steps = num_interp_steps
        self.variables = np.array(VARIABLES)
        self.register_buffer("center", center)
        self.register_buffer("scale", scale)
        self.register_buffer("geop", geop)
        self.register_buffer("lsm", lsm)

    @staticmethod
    def _load_feature_from_file(fn: str, var: str) -> torch.Tensor:
        """Load a feature from a NetCDF file.

        Parameters
        ----------
        fn : str
            Path to the NetCDF file
        var : str
            Variable name to load from the file

        Returns
        -------
        torch.Tensor
            Loaded feature as a tensor with shape (1, 1, H, W)
        """
        with xr.open_dataset(fn) as ds:
            x = np.array(ds[var])
        return torch.Tensor(x).unsqueeze(0).unsqueeze(0)

    def _compute_latlon(self) -> None:
        # compute sin/cos of lat/lon
        coords = self.output_coords(self.input_coords())
        lat = np.deg2rad(coords["lat"])
        lon = np.deg2rad(coords["lon"])
        (lat, lon) = np.meshgrid(lat, lon, indexing="ij")
        sincos_latlon = torch.Tensor(
            np.stack([np.sin(lat), np.cos(lat), np.sin(lon), np.cos(lon)], axis=0)
        ).unsqueeze(0)
        self.register_buffer("lat", torch.as_tensor(lat, device=self.center.device))
        self.register_buffer("lon", torch.as_tensor(lon, device=self.center.device))
        self.register_buffer(
            "sincos_latlon", torch.as_tensor(sincos_latlon, device=self.center.device)
        )

    def __str__(self) -> str:
        return "InterpModAFNO"

    def input_coords(self) -> CoordSystem:
        """Input coordinate system of the prognostic model
        Returns
        -------
        CoordSystem
            Coordinate system dictionary
        """
        # Getter / Setters don't work with torch.nn.Module, need to check manually here
        if self.px_model is None:
            raise ValueError("Base forecast model, px_model, must be set")
        return self.px_model.input_coords()

    @batch_coords()
    def output_coords(self, input_coords: CoordSystem) -> CoordSystem:
        """Output coordinate system of the prognostic model
        Parameters
        ----------
        input_coords : CoordSystem
            Input coordinate system to transform into output_coords
            by default None, will use self.input_coords.
        Returns
        -------
        CoordSystem
            Coordinate system dictionary
        """
        output_coords = OrderedDict(
            {
                "batch": np.empty(0),
                "time": np.empty(0),
                "lead_time": np.array([np.timedelta64(1, "h")]),
                "variable": np.array(self.variables),
                "lat": np.linspace(90.0, -90.0, 720, endpoint=False),
                "lon": np.linspace(0, 360, 1440, endpoint=False),
            }
        )
        if input_coords is None:
            return output_coords
        test_coords = input_coords.copy()
        test_coords["lead_time"] = (
            test_coords["lead_time"] - input_coords["lead_time"][-1]
        )
        # Sometime prognostics don't explicitly have a time coord and thats okay
        # as long as the data does
        output_coords["batch"] = input_coords.get("batch", np.empty(0))
        output_coords["time"] = input_coords.get("time", np.empty(0))
        output_coords["lead_time"] = (
            output_coords["lead_time"] + input_coords["lead_time"]
        )
        return output_coords

    @classmethod
    def load_default_package(cls) -> Package:
        """Load prognostic package"""
        package = Package(
            "ngc://models/nvidia/earth-2/afno_dx_fi-v1-era5@v0.1.0",
            cache_options={
                "cache_storage": Package.default_cache("modafno_interpolation"),
                "same_names": True,
            },
        )
        return package

    @classmethod
    @check_extra_imports("interp-modafno", [PhysicsNemoModule, cos_zenith_angle])
    def load_model(
        cls, package: Package, px_model: PrognosticModel | None = None
    ) -> PrognosticModel:
        """Load prognostic from package

        Parameters
        ----------
        package : Package
            Package to load model from
        px_model : PrognosticModel | None, optional
            The base forecast model that produces the coarse time resolution forecasts.
            If None, should be set manually, by default None

        Returns
        -------
        PrognosticModel
            Prognostic model
        """
        model = PhysicsNemoModule.from_checkpoint(
            package.resolve("fcinterp-modafno-2x2.mdlus")
        )
        model.eval()

        # Load center and std normalizations
        local_center = torch.Tensor(np.load(package.resolve("global_means.npy")))[
            :, :73
        ]
        local_std = torch.Tensor(np.load(package.resolve("global_stds.npy")))[:, :73]

        # load static variables
        geop = cls._load_feature_from_file(package.resolve("orography.nc"), "Z")[
            0, :, :, :720, :
        ]
        geop = (geop - geop.mean()) / geop.std()
        lsm = cls._load_feature_from_file(package.resolve("land_sea_mask.nc"), "LSM")[
            0, :, :, :720, :
        ]

        return cls(
            model,
            center=local_center,
            scale=local_std,
            geop=geop,
            lsm=lsm,
            px_model=px_model,
        )

    def _cos_zenith(self, times: list) -> torch.Tensor:
        """Calculate cosine of zenith angle for given times.

        Parameters
        ----------
        times : list
            List of times to calculate cosine of zenith angle for

        Returns
        -------
        torch.Tensor
            Cosine of zenith angle for each time
        """
        # Convert generator to list to fix type incompatibility
        times_list = list(datetime.fromisoformat(str(t)[:19]) for t in times)
        cos_zen = [
            cos_zenith_angle(t, self.lon.cpu().numpy(), self.lat.cpu().numpy())
            for t in times_list
        ]
        return torch.Tensor(np.stack(cos_zen, axis=0)).unsqueeze(0)

    @torch.inference_mode()
    def _interpolate(
        self,
        x0: torch.Tensor,
        x1: torch.Tensor,
        coords: CoordSystem,
    ) -> Generator[tuple[torch.Tensor, CoordSystem], None, None]:
        """Interpolate between two forecast steps.

        Parameters
        ----------
        x0 : torch.Tensor
            First forecast step
        x1 : torch.Tensor
            Second forecast step
        coords : CoordSystem
            Coordinate system for the forecast steps

        Yields
        ------
        Generator[tuple[torch.Tensor, CoordSystem], None, None]
            Generator yielding interpolated forecast steps and their coordinate systems
        """
        x0 = (x0 - self.center) / self.scale
        x1 = (x1 - self.center) / self.scale

        out = torch.zeros_like(x0)
        batch_size = len(coords["batch"])

        # expand constants
        sincos_latlon = self.sincos_latlon.expand(batch_size, -1, -1, -1)
        geop = self.geop.expand(batch_size, -1, -1, -1)
        lsm = self.lsm.expand(batch_size, -1, -1, -1)

        t0 = coords["time"][:, None] + coords["lead_time"][None, :]
        coords_end = coords
        for _ in range(self.num_interp_steps):
            coords_end = self.output_coords(coords_end)
        t1 = coords_end["time"][:, None] + coords_end["lead_time"][None, :]

        for interp_step in range(1, self.num_interp_steps):
            coords = self.output_coords(coords)

            for ti, t in enumerate(coords["time"]):
                for lti, lt in enumerate(coords["lead_time"]):
                    t_ip = t + lt
                    cos_zen = self._cos_zenith([t0[ti, lti], t1[ti, lti], t_ip]).to(
                        device=x0.device
                    )

                    x = torch.concat(
                        [
                            x0[:, ti, lti],
                            x1[:, ti, lti],
                            cos_zen.expand(batch_size, -1, -1, -1),
                            sincos_latlon,
                            geop,
                            lsm,
                        ],
                        dim=1,
                    )

                    t_norm = torch.Tensor([interp_step / self.num_interp_steps]).to(
                        device=x0.device
                    )
                    out[:, ti, lti] = self.interp_model(x, t_norm)

            out *= self.scale
            out += self.center
            yield (out, coords)

    @batch_func()
    def __call__(
        self,
        x: torch.Tensor,
        coords: CoordSystem,
    ) -> tuple[torch.Tensor, CoordSystem]:
        """Runs prognostic model 1 step

        Parameters
        ----------
        x : torch.Tensor
            Input tensor
        coords : CoordSystem
            Input coordinate system

        Returns
        -------
        tuple[torch.Tensor, CoordSystem]
            Output tensor and coordinate system 1 hour in the future
        """
        gen = self._default_generator(x, coords)
        return next(gen)

    @batch_func()
    def _default_generator(
        self, x: torch.Tensor, coords: CoordSystem
    ) -> Generator[tuple[torch.Tensor, CoordSystem], None, None]:

        if self.px_model is None:
            raise ValueError(
                "Base forecast model, px_model, must be set before executing the model."
            )

        if not hasattr(self, "sincos_latlon"):
            self._compute_latlon()

        for fc_step, (x, coords) in enumerate(self.px_model.create_iterator(x, coords)):
            # Make sure prognostic model has all 73 required variables
            (x, coords) = map_coords(x, coords, self.output_coords(coords))
            if fc_step == 0:
                x0 = x
                coords0 = coords
                yield (x, coords)
            else:
                x1 = x
                for x, coords in self._interpolate(x0, x1, coords0):
                    yield (x, coords)
                coords = self.output_coords(coords)
                yield (x1, coords)
                x0 = x1
                coords0 = coords

    def create_iterator(
        self, x: torch.Tensor, coords: CoordSystem
    ) -> Iterator[tuple[torch.Tensor, CoordSystem]]:
        """Creates a iterator which can be used to perform time-integration of the
        prognostic model. Will return the initial condition first (0th step).

        Parameters
        ----------
        x : torch.Tensor
            Input tensor
        coords : CoordSystem
            Input coordinate system

        Yields
        ------
        Iterator[tuple[torch.Tensor, CoordSystem]]
            Iterator that generates time-steps of the prognostic model container the
            output data tensor and coordinate system dictionary.
        """
        yield from self._default_generator(x, coords)



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

try:
    import cupy as cp
    from cucim.skimage.feature import peak_local_max as cucim_peak_local_max
    from cucim.skimage.measure import label, regionprops
    from cucim.skimage.morphology import binary_erosion, remove_small_objects
    from scipy.spatial import KDTree
    from skimage.feature import peak_local_max as skimage_peak_local_max
    from skimage.morphology import convex_hull_image

    # from cupyx.scipy.spatial import KDTree as CuKDTree
except ImportError:
    cp = None
    cucim_peak_local_max = None
    label = None
    regionprops = None
    binary_erosion = None
    remove_small_objects = None
    skimage_peak_local_max = None
    convex_hull_image = None
    KDTree = None
    # CuKDTree = None

import numpy as np
import torch

from earth2studio.models.batch import batch_coords, batch_func
from earth2studio.utils import (
    handshake_coords,
    handshake_dim,
)
from earth2studio.utils.imports import check_extra_imports
from earth2studio.utils.type import CoordSystem

VARIABLES_TCV = [
    "u10m",
    "v10m",
    "msl",
    "u850",
    "v850",
    "z500",
    "z850",
    "z200",
    "t500",
    "t400",
    "t300",
    "t250",
    "t200",
]
VARIABLES_TCWD = ["u10m", "v10m", "msl", "u850", "v850"]
OUT_VARIABLES = ["tc_lat", "tc_lon", "tc_msl", "tc_w10m"]


class _TCTrackerBase:

    PATH_FILL_VALUE = -9999  # Should not be in lat/lon range for safety

    @classmethod
    def vorticity(
        cls, u: torch.Tensor, v: torch.Tensor, dx: float = 25000.0, dy: float = 25000.0
    ) -> torch.Tensor:
        """Compute Relative Vorticity."""
        dudx = torch.gradient(u, dim=-2)[0] / dx
        dvdy = torch.gradient(v, dim=-1)[0] / dy
        vorticity = torch.add(dvdy, dudx)
        return vorticity

    @classmethod
    def haversine_torch(
        cls,
        lat1: torch.Tensor,
        lon1: torch.Tensor,
        lat2: torch.Tensor,
        lon2: torch.Tensor,
    ) -> torch.Tensor:
        """Compute haversine distance between two pairs of lat/lon points in km.

        Parameters
        ----------
        lat1 : torch.Tensor
            Latitude coordinates of first point [n]
        lon1 : torch.Tensor
            Longitude coordinates of first point [n]
        lat2 : torch.Tensor
            Latitude coordinates of second point [n]
        lon2 : torch.Tensor
            Longitude coordinates of second point [n]

        Returns
        -------
        torch.Tensor
            Distance between two points [n]
        """
        lat1, lon1, lat2, lon2 = map(torch.deg2rad, [lat1, lon1, lat2, lon2])
        dlon = lon2 - lon1
        dlat = lat2 - lat1
        unit = 6371
        return (
            2
            * unit
            * torch.arcsin(
                torch.sqrt(
                    torch.sin(dlat / 2) ** 2
                    + torch.cos(lat1) * torch.cos(lat2) * torch.sin(dlon / 2) ** 2
                )
            )
        )

    @staticmethod
    def latlon_to_equirectangular(
        lat: torch.Tensor, lon: torch.Tensor, R: float = 6371.0
    ) -> torch.Tensor:
        """Convert latitude/longitude to equirectangular projection coordinates.

        Parameters
        ----------
        lat : torch.Tensor
            Latitude in degrees
        lon : torch.Tensor
            Longitude in degrees
        R : float, optional
            Earth radius in km, by default 6371.0

        Returns
        -------
        torch.Tensor
            Stacked x, y coordinates in km
        """
        lat_rad = torch.deg2rad(lat)
        lon_rad = torch.deg2rad(lon)

        # Project to equirectangular coordinates
        x = R * lon_rad * torch.cos(torch.deg2rad(torch.tensor(0.0)))
        y = R * lat_rad

        return torch.stack([x, y], dim=-1)

    @classmethod
    def get_local_max(
        cls,
        x: torch.Tensor,
        threshold_abs: float | None = None,
        min_distance: int = 1,
        exclude_border: bool | int = True,
    ) -> torch.Tensor:
        """Gets the local maximum of a tensor x, above a given absolute threshold,
        with a minimum distance separating local maximums.

        This is a helper utility that converts a pytorch tensor to a cupy tensor
        to use a CuCIM utility `peak_local_max` to extract the local maxima.

        Note
        ----
        For more details, see:

        - https://scikit-image.org/docs/stable/api/skimage.feature.html#skimage.feature.peak_local_max


        Parameters
        ----------
        x : torch.Tensor
            Input tensor of shape [x, y]
        threshold_abs : _type_, optional
            Absolute value to threshold local maximum, by default None
        min_distance : int, optional
            Minimum distance separating local maximums, by default 1
        exclude_border: bool or int
            If positive integer, exclude_border excludes peaks from within
            exclude_border-pixels of the border of the image. If tuple of
            non-negative ints, the length of the tuple must match the input
            array dimensionality. Each element of the tuple will exclude peaks
            from within exclude_border-pixels of the border of the image along
            that dimension. If True, takes the min_distance parameter as value.
            If zero or False, peaks are identified regardless of their distance
            from the border.

        Returns
        -------
        torch.Tensor
            List of coordinates of local maximum [2, N]
        """
        if x.is_cuda:
            x_ = cp.from_dlpack(x)
            local_max = cucim_peak_local_max(
                x_,
                threshold_abs=threshold_abs,
                min_distance=min_distance,
                exclude_border=exclude_border,
            )
            local_max = torch.from_dlpack(local_max)
        else:
            x_ = np.from_dlpack(x)
            local_max = skimage_peak_local_max(
                x_,
                threshold_abs=threshold_abs,
                min_distance=min_distance,
                exclude_border=exclude_border,
            )
            local_max = torch.as_tensor(local_max, device=x.device)

        return local_max

    @classmethod
    def append_paths(
        cls,
        frame: torch.Tensor,
        path_buffer: torch.Tensor,
        path_search_distance: float = 250,
        path_search_window_size: int = 3,
    ) -> torch.Tensor:
        """Appends frame of TC centers into the track path history tensor

        Parameters
        ----------
        frame : torch.Tensor
            Instanteous frame of TC centers to append to existing list of paths of size
            [batch, point_id, variable]
        path_buffer : torch.Tensor
            The current buffer of paths to append to with size
            [batch, path_id, step, variable]
        path_search_distance : float, optional
            Max haversine search distance to connect to in km, by default 250
        path_search_window_size: int, optional
            The historical window size for a path to use when pairing. Namely, the
            path search will use the specified number of historic points to connect a
            new frame to the current set of paths, by default 3

        Returns
        -------
        torch.Tensor
            Updated history buffer of size [batch, path_id(+1), step+1, variable]
        """
        if path_buffer.nelement() == 0:
            return frame.unsqueeze(2).clone()

        if path_search_window_size < 1:
            raise ValueError("Path search window size must be greater than 1")

        if path_search_distance <= 0:
            raise ValueError("Path search distance must be greater than 0")

        if frame.shape[0] != path_buffer.shape[0]:
            raise ValueError(
                f"Error with updating TC tracker history buffer, input and history buffer need the same batch size. Got {frame.shape[0]} and {path_buffer.shape[0]}"
            )

        next_frame = torch.full_like(
            path_buffer[:, :, 0, :], cls.PATH_FILL_VALUE, dtype=frame.dtype
        ).unsqueeze(2)
        # Expand by one path for over flow / new paths
        # I.e. we have [batch, path id + 1, 1, variable]
        next_frame = torch.cat(
            [next_frame, torch.full_like(next_frame[:, :1], cls.PATH_FILL_VALUE)], dim=1
        )

        for i in range(frame.shape[0]):
            # Stack all lat/lon features for window size into a [n,2] feature list
            path_search_window_size = min(
                [path_search_window_size, path_buffer.shape[2]]
            )
            tree_features = (
                path_buffer[i, :, -path_search_window_size:, :2].flip(-2).reshape(-1, 2)
            )

            tree_features = cls.latlon_to_equirectangular(
                tree_features[:, 0], tree_features[:, 1]
            )
            input_features = cls.latlon_to_equirectangular(
                frame[i, :, 0], frame[i, :, 1]
            )

            if tree_features.is_cuda and False:
                # Make need to edit below, not worth it at the moment
                # /.venv/lib/python3.12/site-packages/cupy/_environment.py
                # Liune 350 to min_pypi_version = config[lib]['version']
                # tree = CuKDTree(cp.from_dlpack(tree_features))
                # dist, idx = tree.query(cp.from_dlpack(input_features), k=1)
                pass
            else:
                tree = KDTree(tree_features.cpu())
                dist, idx = tree.query(input_features.cpu(), k=1)

            for p in range(dist.shape[0]):
                # If all variables are 0, its a filler from rnn.pad_sequence
                if torch.all(frame[i, p] == cls.PATH_FILL_VALUE):
                    continue
                # For steps that are further back in the window, increase radius
                # Note, the `.flip(-2)` above make this calculation easier placing most
                # recent step at index 0 instead of last
                past_steps = idx[p] % path_search_window_size + 1
                if dist[p] < past_steps * path_search_distance:
                    # Recall we have a window of points for each path
                    # if any match, append to that path
                    path_id = int(idx[p]) // path_search_window_size
                    next_frame[i, path_id] = frame[i, p]
                # No match so looks like we need a new path
                else:
                    next_frame[i, -1] = frame[i, p]

        # if theres nothing in the extra path row, get rid of it
        if torch.all(next_frame[:, -1] == cls.PATH_FILL_VALUE):
            next_frame = next_frame[:, :-1]
        else:
            # Expand the path_id dim by 1 for concat
            import torch.nn.functional as F

            path_buffer = F.pad(
                path_buffer, (0, 0, 0, 0, 0, 1, 0, 0), "constant", cls.PATH_FILL_VALUE
            )
        return torch.cat([path_buffer, next_frame], axis=2)


@check_extra_imports("cyclone", [cp, KDTree, "cucim", "skimage"])
class TCTrackerWuDuan(torch.nn.Module, _TCTrackerBase):
    """Finds a list of tropical cyclone (TC) centers using an adaption of the method
    described in the conditions in Wu and Duan 2023. The algorithm converts vorticity
    from reanalysis data into a binary image using a defined critical threshold.
    Subsequent processing with connected component labeling and erosion identifies the
    resulting inner cores as TC seeds.

    Note
    ----
    For more information about this method see:

    - https://doi.org/10.1016/j.wace.2023.100626


    Parameters
    ----------
    path_search_distance: int, optional
        The max radial distance two cyclone centers will be considered part of the same
        path in km, by default 300
    path_search_window_size: int, optional
        The historical window size used when creating TC paths, by default 2

    Examples
    --------
    The cyclone tracker will return a tensor of TC paths collected over a series of
    forward passes which are held inside of the models state.
    Namely given a time series of `n` snap shots, the tracker should be called for each
    time-step resulting in a tensor consisting of a set number of paths with n steps.
    Any non-valid / missing data will be `torch.nan` for filtering in post processing
    steps.

    >>> model = TCTrackerWuDuan()
    >>> # Process each timestep
    >>> for time in [datetime(2017, 8, 25) + timedelta(hours=6 * i) for i in range(3)]:
    ...     da = data_source(time, tracker.input_coords()["variable"])
    ...     input, input_coords = prep_data_array(da, device=device)
    ...     output, output_coords = model(input, input_coords)
    >>> # Final path_buffer shape: [batch, path_id, steps, variable]
    >>> output.shape  # torch.Size([1, 6, 3, 4])
    >>> model.path_buffer.shape  # torch.Size([1, 6, 3, 4])
    >>> # Remove current paths from models state
    >>> model.reset_path_buffer()
    >>> model.path_buffer.shape  # torch.Size([0])
    """

    def __init__(
        self, path_search_distance: int = 300, path_search_window_size: int = 2
    ) -> None:
        super().__init__()
        self.register_buffer("path_buffer", torch.empty(0))
        self.path_search_distance = path_search_distance
        self.path_search_window_size = path_search_window_size

    def reset_path_buffer(self) -> None:
        """Resets the internal"""
        self.path_buffer = torch.empty(0)

    def input_coords(self) -> CoordSystem:
        """Input coordinate system of diagnostic model

        Returns
        -------
        CoordSystem
            Coordinate system dictionary
        """
        return OrderedDict(
            {
                "batch": np.empty(0),
                "variable": np.array(VARIABLES_TCWD),
                "lat": np.linspace(90, -90, 721, endpoint=True),
                "lon": np.linspace(0, 360, 1440, endpoint=False),
            }
        )

    @batch_coords()
    def output_coords(self, input_coords: CoordSystem) -> CoordSystem:
        """Output coordinate system of diagnostic model

        Parameters
        ----------
        input_coords : CoordSystem
            Input coordinate system to transform into output_coords
            by default None, will use self.input_coords.

        Returns
        -------
        CoordSystem
            Coordinate system dictionary
        """
        target_input_coords = self.input_coords()
        handshake_dim(input_coords, "lon", 3)
        handshake_dim(input_coords, "lat", 2)
        handshake_dim(input_coords, "variable", 1)
        handshake_coords(input_coords, target_input_coords, "lon")
        handshake_coords(input_coords, target_input_coords, "lat")
        handshake_coords(input_coords, target_input_coords, "variable")

        # [batch, path_id, step, variable]
        output_coords = OrderedDict(
            [
                ("batch", input_coords["batch"]),
                ("path_id", np.empty(0)),
                ("step", np.empty(0)),
                ("variable", np.array(OUT_VARIABLES)),
            ]
        )
        return output_coords

    def _find_centers(
        self,
        lat: torch.Tensor,
        lon: torch.Tensor,
        vort850: torch.Tensor,
        w10m: torch.Tensor,
        msl: torch.Tensor,
        vort850_threshold: torch.Tensor = torch.tensor(1.4e-4),
    ) -> torch.Tensor:
        """Finds a list of tropical cyclone centers

        Parameters
        ----------
        lat : torch.Tensor
            Vector of latitudes for tensors.
        lon : torch.Tensor
            Vector of longitudes for tensors.
        vort850 : torch.Tensor
            850 hPa relative vorticity of dimension [lat, lon].

        Returns
        -------
        torch.Tensor
            List of TC centers, torch.Tensor of shape [N, 4]
        """

        if vort850.is_cuda:
            v_ = cp.from_dlpack(vort850)
        else:
            v_ = cp.array(np.from_dlpack(vort850))
        if w10m.is_cuda:
            w10m_ = cp.from_dlpack(w10m)
        else:
            w10m_ = cp.array(np.from_dlpack(w10m))
        if msl.is_cuda:
            msl_ = cp.from_dlpack(msl)
        else:
            msl_ = cp.array(np.from_dlpack(msl))
        vort850_threshold = float(vort850_threshold)
        x_ = v_ > vort850_threshold
        # Label regions
        x_label = label(x_)

        # Remove labels that are not of the the needed size (less than 18 pixels)
        x_label = remove_small_objects(x_label, min_size=18, connectivity=1)
        # Get region props (bounding box / axis)
        props = regionprops(x_label, intensity_image=v_)
        centers = []
        for prop in props:
            if prop.axis_minor_length == 0:
                # This detected object is too slim to be a tropical storm
                continue

            object_ratio = prop.axis_major_length / prop.axis_minor_length
            if object_ratio > 2:
                # This object is not round enough
                continue

            # extend prop slice to get on each side of the object an additional line of pixels
            start = max(0, prop.slice[0].start - 1)
            stop = min(x_label.shape[0], prop.slice[0].stop + 1)
            x_slice_ext = slice(start, stop, prop.slice[0].step)
            start = max(0, prop.slice[1].start - 1)
            stop = min(x_label.shape[1], prop.slice[1].stop + 1)
            y_slice_ext = slice(start, stop, prop.slice[1].step)
            prop_slice_ext = (x_slice_ext, y_slice_ext)

            # calculate exact center of storm in longitude and latitude coordinates
            # weighted centroid position of center pixels are used (weighted by vorticity)
            lat_idx, lon_idx = prop.centroid_weighted
            grid_spacing = lon[int(cp.ceil(lon_idx))] - lon[int(cp.floor(lon_idx))]
            residual_step = float(lon_idx - cp.floor(lon_idx))
            lon_ = lon[int(cp.floor(lon_idx))] + residual_step * grid_spacing
            grid_spacing = lat[int(cp.ceil(lat_idx))] - lat[int(cp.floor(lat_idx))]
            residual_step = float(lat_idx - cp.floor(lat_idx))
            lat_ = lat[int(cp.floor(lat_idx))] + residual_step * grid_spacing

            # erode object to get center of storm
            bool_center = binary_erosion(x_label[prop_slice_ext]).astype(bool)

            # skip object if the center is smaller or equal 1 pixel
            if bool_center.sum() <= 1:
                continue

            # check solidity of storm center (eroded object) and storm (uneroded object)
            solidity_center = (
                bool_center.sum() / convex_hull_image(bool_center.get()).sum()
            )
            solidity_storm = prop.solidity
            # objects that are not solid have many holes in the vorticity field are not a tropical storm
            if solidity_center < 0.3:
                continue
            if solidity_storm < 0.3:
                continue

            # a storm needs to be solid (solidity) and round (object ratio)
            if object_ratio / prop.solidity > 4:
                continue

            # skip objects that are not in the tropical storm areas
            # too close to the equator
            if abs(lat_) < 5:
                continue
            if not (lon_ > 20 and lat_ > 5 and lat_ < 50) and not (
                lon_ < 200 and lat_ < -5 and lat_ > -40
            ):
                continue

            # define a mask that describes the pixels that sourrand the center/the pixels that were removed by binary_erosion
            bool_border = x_label[prop_slice_ext].astype(bool) & ~bool_center
            from copy import copy

            # calculate the mean of the voritcity of the border pixels and the center pixels
            v_sub1 = copy(v_[prop_slice_ext])
            v_sub2 = copy(v_sub1)
            cp.putmask(v_sub1, ~bool_center, cp.nan)
            cp.putmask(v_sub2, ~bool_border, cp.nan)
            mean_inner = cp.nanmean(v_sub1)
            mean_outer = cp.nanmean(v_sub2)

            # mean voriticty in the center must be larger than mean vorticity at the border of the storm
            if mean_inner > mean_outer:
                # calculate max wind and lowest msl
                w10m_max = torch.tensor(w10m_[prop_slice_ext].max())
                msl_min = torch.tensor(msl_[prop_slice_ext].min())

                # add storm center to list of centers
                centers.append(
                    torch.as_tensor(
                        [lat_, lon_, msl_min, w10m_max], device=vort850.device
                    )
                )

        if len(centers) > 0:
            return torch.stack(centers, dim=0)
        else:
            x = torch.full((1, 4), self.PATH_FILL_VALUE, device=vort850.device)
            return x

    @torch.inference_mode()
    @batch_func()
    def __call__(
        self,
        x: torch.Tensor,
        coords: CoordSystem,
    ) -> tuple[torch.Tensor, CoordSystem]:
        """Forward pass of diagnostic"""

        output_coords = self.output_coords(coords)

        lat = torch.as_tensor(self.input_coords()["lat"], device=x.device)
        lon = torch.as_tensor(self.input_coords()["lon"], device=x.device)

        def get_variable(x: torch.Tensor, var: str) -> torch.Tensor:
            index = VARIABLES_TCWD.index(var)
            return x[index]

        outs = []
        for i in range(x.shape[0]):

            # Get wind components at 850 hPa
            # VARIABLES_TCWD = ["u850", "v850", "u10m", "v10m", "msl"]
            u850 = get_variable(x[i], "u850")
            v850 = get_variable(x[i], "v850")
            u10m = get_variable(x[i], "u10m")
            v10m = get_variable(x[i], "v10m")
            msl = get_variable(x[i], "msl")

            # Calculate vorticity at 850 hPa
            vort850 = TCTrackerWuDuan.vorticity(u850, v850)
            vort850[361:] *= -1  # Invert southern hemisphere

            # Calculate wind speed at 10m height
            w10m = torch.sqrt(torch.square(u10m) + torch.square(v10m))

            # identify position of tropical storm centers
            centers = self._find_centers(lat, lon, vort850, w10m, msl)
            # Pack [points, variables]
            outs.append(centers)

        # amazing function!
        # https://pytorch.org/docs/stable/generated/torch.nn.utils.rnn.pad_sequence.html
        out = torch.nn.utils.rnn.pad_sequence(
            outs, padding_value=self.PATH_FILL_VALUE, batch_first=True
        )
        # [batch, path_id, step, variable]
        self.path_buffer = self.append_paths(
            out,
            self.path_buffer,
            self.path_search_distance,
            self.path_search_window_size,
        )
        out = torch.where(
            self.path_buffer == self.PATH_FILL_VALUE, torch.nan, self.path_buffer
        )

        output_coords = self.output_coords(coords)
        output_coords["path_id"] = np.arange(self.path_buffer.shape[1])
        output_coords["step"] = np.arange(self.path_buffer.shape[2])

        return out, output_coords


@check_extra_imports("cyclone", [cp, KDTree, "cucim", "skimage"])
class TCTrackerVitart(torch.nn.Module, _TCTrackerBase):
    """Finds a list of tropical cyclone centers using the conditions in Vitart 1997

    Note
    ----
    For more information about this method see:

    - https://doi.org/10.1175/1520-0442(1997)010%3C0745:SOIVOT%3E2.0.CO;2


    Parameters
    ----------
    vorticity_threshold : float, optional
        The threshold for vorticity at 850, below which a possible
        tropical cyclone center is rejected, by default 3.5e-5 1/s
    mslp_threshold : float, optional
        The threshold for minimum sea level pressure for local minimums
        to be considered tropical cyclone, by default 99000 Pa
    temp_dec_threshold : float, optional
        The value for which average temperature must decrease away from
        the warm core for a possible center to be considered a tropical
        cyclone, by default 0.5 degrees celsius
    lat_threshold: float, optional
        The maximum absolute latitude that a point will be considered to
        be a tropical cyclone, by default 60 degrees (N and S).
    exclude_border: bool | int, optional
        If positive integer, exclude_border excludes peaks from within
        exclude_border-pixels of the border of the image. If tuple of
        non-negative ints, the length of the tuple must match the input
        array dimensionality. Each element of the tuple will exclude peaks
        from within exclude_border-pixels of the border of the image along
        that dimension. If True, takes the min_distance parameter as value.
        If zero or False, peaks are identified regardless of their distance
        from the border.
    path_search_distance: int, optional
        The max radial distance two cyclone centers will be considered part of the same
        path in km, by default 300
    path_search_window_size: int, optional
        The historical window size used when creating TC paths, by default 2

    Examples
    --------
    The cyclone tracker will return a tensor of TC paths collected over a series of
    forward passes which are held inside of the models state.
    Namely given a time series of `n` snap shots, the tracker should be called for each
    time-step resulting in a tensor consisting of a set number of paths with n steps.
    Any non-valid / missing data will be `torch.nan` for filtering in post processing
    steps.

    >>> model = TCTrackerVitart()
    >>> # Process each timestep
    >>> for time in [datetime(2017, 8, 25) + timedelta(hours=6 * i) for i in range(3)]:
    ...     da = data_source(time, tracker.input_coords()["variable"])
    ...     input, input_coords = prep_data_array(da, device=device)
    ...     output, output_coords = model(input, input_coords)
    >>> # Final path_buffer shape: [batch, path_id, steps, variable]
    >>> output.shape  # torch.Size([1, 6, 3, 4])
    >>> model.path_buffer.shape  # torch.Size([1, 6, 3, 4])
    >>> # Remove current paths from models state
    >>> model.reset_path_buffer()
    >>> model.path_buffer.shape  # torch.Size([0])
    """

    def __init__(
        self,
        vorticity_threshold: float = 3.5e-5,
        mslp_threshold: float = 99000.0,
        temp_dec_threshold: float = 0.5,
        lat_threshold: float = 60.0,
        exclude_border: bool | int = True,
        path_search_distance: int = 300,
        path_search_window_size: int = 2,
    ) -> None:
        super().__init__()
        # TC Center identification parameters
        self.vorticity_threshold = vorticity_threshold
        self.msl_threshold = mslp_threshold
        self.temp_dec_threshold = temp_dec_threshold
        self.lat_threshold = lat_threshold
        self.exclude_border = exclude_border
        # TC path identification parameters
        self.register_buffer("path_buffer", torch.empty(0))
        self.path_search_distance = path_search_distance
        self.path_search_window_size = path_search_window_size

    def reset_path_buffer(self) -> None:
        """Resets the internal"""
        self.path_buffer = torch.empty(0)

    def input_coords(self) -> CoordSystem:
        """Input coordinate system of diagnostic model

        Returns
        -------
        CoordSystem
            Coordinate system dictionary
        """
        return OrderedDict(
            {
                "batch": np.empty(0),
                "variable": np.array(VARIABLES_TCV),
                "lat": np.linspace(90, -90, 721, endpoint=True),
                "lon": np.linspace(0, 360, 1440, endpoint=False),
            }
        )

    @batch_coords()
    def output_coords(self, input_coords: CoordSystem) -> CoordSystem:
        """Output coordinate system of diagnostic model

        Parameters
        ----------
        input_coords : CoordSystem
            Input coordinate system to transform into output_coords
            by default None, will use self.input_coords.

        Returns
        -------
        CoordSystem
            Coordinate system dictionary
        """
        target_input_coords = self.input_coords()
        handshake_dim(input_coords, "lon", 3)
        handshake_dim(input_coords, "lat", 2)
        handshake_dim(input_coords, "variable", 1)
        handshake_coords(input_coords, target_input_coords, "variable")

        # [batch, path_id, step, variable]
        output_coords = OrderedDict(
            [
                ("batch", input_coords["batch"]),
                ("path_id", np.empty(0)),
                ("step", np.empty(0)),
                ("variable", np.array(OUT_VARIABLES)),
            ]
        )
        return output_coords

    def _find_centers(
        self,
        lat: torch.Tensor,
        lon: torch.Tensor,
        msl: torch.Tensor,
        w10m: torch.Tensor,
        vort850: torch.Tensor,
        t_200_500_mean: torch.Tensor,
        dz_200_850: torch.Tensor,
        vorticity_threshold: float = 3.5e-5,
        mslp_threshold: float = 99000.0,
        temp_dec_threshold: float = 0.5,
        exclude_border: bool | int = True,
    ) -> torch.Tensor:
        """Finds a list of tropical cyclone centers

        Parameters
        ----------
        lat : torch.Tensor
            mesh of latitudes tensors. [nlat, nlon]
        lon : torch.Tensor
            mesh of longitudes tensors. [nlat, nlon]
        msl : torch.Tensor
            Mean Sea Level Pressure tensor of dimension [nlat, nlon].
        w10m : torch.Tensor
            Surface wind tensor, just used to return surface wind at core
        vort850 : torch.Tensor
        850 hPa relative vorticity of dimension [nlat, nlon].
        t_200_500_mean : torch.Tensor
            Average temperature between 200 hPa and 500 hPa of
            dimension [nlat, nlon].
        dz_200_850 : torch.Tensor
            Thickness between 200 hPa and 850 hPa geopotential. [nlat, nlon]
        vorticity_threshold: float
            The threshold for vorticity at 850, below which a possible
            tropical cyclone center is rejected. By default 3.5e-5 1/s
        mslp_threshold: float
            The threshold for minimum sea level pressure for local minimums
            to be considered tropical cyclones. By default 990000 Pa.
        temp_dec_threshold: float
            The value for which average temperature must decrease away from
            the warm core for a possible center to be considered a tropical
            cyclone. By default 0.5 degrees celsius
        exclude_border: bool or int
            If positive integer, exclude_border excludes peaks from within
            exclude_border-pixels of the border of the image. If tuple of
            non-negative ints, the length of the tuple must match the input
            array dimensionality. Each element of the tuple will exclude peaks
            from within exclude_border-pixels of the border of the image along
            that dimension. If True, takes the min_distance parameter as value.
            If zero or False, peaks are identified regardless of their distance
            from the border.

        Returns
        -------
        torch.Tensor
            List of TC centers, torch.Tensor of shape [2, N]
        """

        min_distance = 10
        # Get local max vorticity
        vlm = TCTrackerVitart.get_local_max(
            vort850,
            threshold_abs=vorticity_threshold,
            min_distance=min_distance,
            exclude_border=exclude_border,
        )
        vlm_loc = torch.stack(
            (lat[vlm[:, 0], vlm[:, 1]], lon[vlm[:, 0], vlm[:, 1]]), dim=1
        )

        # Get local maximum of average temperature between 500 and 200mb
        tlm = TCTrackerVitart.get_local_max(
            t_200_500_mean, min_distance=min_distance, exclude_border=exclude_border
        )
        tlm_loc = torch.stack(
            (lat[tlm[:, 0], tlm[:, 1]], lon[tlm[:, 0], tlm[:, 1]]), dim=1
        )

        # Get local max z200 - z850
        dzlm = TCTrackerVitart.get_local_max(dz_200_850, exclude_border=exclude_border)
        dzlm_loc = torch.stack(
            (lat[dzlm[:, 0], dzlm[:, 1]], lon[dzlm[:, 0], dzlm[:, 1]]), dim=1
        )

        # Get local min msl
        msllm = TCTrackerVitart.get_local_max(
            -msl,
            threshold_abs=-mslp_threshold,
            min_distance=min_distance,
            exclude_border=exclude_border,
        )
        mlm_loc = torch.stack(
            (lat[msllm[:, 0], msllm[:, 1]], lon[msllm[:, 0], msllm[:, 1]]), dim=1
        )

        dzlm = TCTrackerVitart.get_local_max(dz_200_850, exclude_border=exclude_border)
        dzlm_loc = torch.stack(
            (lat[dzlm[:, 0], dzlm[:, 1]], lon[dzlm[:, 0], dzlm[:, 1]]), dim=1
        )

        centers = []
        for i, mins in enumerate(mlm_loc):
            idx_lat, idx_lon = msllm[i]
            center0 = torch.as_tensor(mins, device=msl.device)

            # Vorticity filter
            dist = TCTrackerVitart.haversine_torch(
                mins[0], mins[1], vlm_loc[:, 0], vlm_loc[:, 1]
            )
            if dist.min() > 8 * 25:  # Distance should be 8 degrees ~= 200 km
                continue

            # Warm Core filter
            dist = TCTrackerVitart.haversine_torch(
                mins[0], mins[1], tlm_loc[:, 0], tlm_loc[:, 1]
            )
            if dist.min() > 2 * 25:  # Distance should be less than 2 degrees ~= 50 km
                continue

            ti = torch.argmin(dist)  # Get closest distance to core
            ti_loc = tlm_loc[ti]
            ti = tlm[ti]

            lat_inds = (
                TCTrackerVitart.haversine_torch(
                    ti_loc[0], ti_loc[1], lat[:, ti[1]], lon[:, ti[1]]
                )
                < 8 * 25
            )
            lon_inds = (
                TCTrackerVitart.haversine_torch(
                    ti_loc[0], ti_loc[1], lat[ti[0], :], lon[ti[0], :]
                )
                < 8 * 25
            )
            dec = (
                t_200_500_mean[:, lon_inds][lat_inds]  # approximation to stay regular
                - t_200_500_mean[ti[0], ti[1]]
            ) < -temp_dec_threshold
            new_ti_lat_index = torch.where(lat[lat_inds, ti[1]] == ti_loc[0])[0].item()
            new_ti_lon_index = torch.where(lon[ti[0], lon_inds] == ti_loc[1])[0].item()
            if not all(
                [
                    torch.any(dec[new_ti_lat_index, new_ti_lon_index:]),
                    torch.any(dec[new_ti_lat_index, :new_ti_lon_index]),
                    torch.any(dec[:new_ti_lat_index, new_ti_lon_index]),
                    torch.any(dec[new_ti_lat_index:, new_ti_lon_index]),
                ]
            ):
                continue

            # dZ filter
            dist = TCTrackerVitart.haversine_torch(
                mins[0], mins[1], dzlm_loc[:, 0], dzlm_loc[:, 1]
            )
            if dist.min() > 2 * 25:
                continue

            w10m_max = torch.max(
                w10m[
                    idx_lat - min_distance : idx_lat + min_distance,
                    idx_lon - min_distance : idx_lon + min_distance,
                ]
            )
            centers.append(
                torch.tensor([center0[0], center0[1], msl[ti[0], ti[1]], w10m_max]).to(
                    msl.device
                )
            )

        if len(centers) > 0:
            return torch.stack(centers, dim=0)
        else:
            x = torch.full((1, 4), self.PATH_FILL_VALUE, device=vort850.device)
            return x

    @torch.inference_mode()
    @batch_func()
    def __call__(
        self,
        x: torch.Tensor,
        coords: CoordSystem,
    ) -> tuple[torch.Tensor, CoordSystem]:
        """Forward pass of diagnostic"""

        output_coords = self.output_coords(coords)

        lat = torch.as_tensor(coords["lat"], device=x.device)
        lon = torch.as_tensor(coords["lon"], device=x.device)

        if lat.ndim != lon.ndim:
            raise ValueError(
                "Error, lat/lon grids must have the same number of dimensions."
            )

        if lat.ndim < 2:
            lon, lat = torch.meshgrid(lon, lat, indexing="xy")

        if lat.shape != lon.shape:
            raise ValueError("Error, lat/lon grids must be the same shape.")

        if self.lat_threshold is not None:
            _, nlon = lat.shape
            indices = lat.abs() < self.lat_threshold

            lat = lat[indices].reshape(-1, nlon)
            lon = lon[indices].reshape(-1, nlon)

        def get_variable(x0: torch.Tensor, var: str) -> torch.Tensor:
            index = VARIABLES_TCV.index(var)
            return x0[:, index]

        ####
        # First thing to do is to get MSL local minimums
        # extract MSL from x
        # x - [n, 8, nlat, nlon]

        u10m = get_variable(x, "u10m")
        v10m = get_variable(x, "v10m")
        w10m = torch.sqrt(torch.pow(u10m, 2) + torch.pow(v10m, 2))

        # Get vorticity
        u850 = get_variable(x, "u850")
        v850 = get_variable(x, "v850")
        vort850 = TCTrackerVitart.vorticity(u850, v850)
        vort850[:, 361:] *= -1
        # Get MSL
        msl = get_variable(x, "msl")
        # Get average temp
        t_200_500_mean = torch.mean(
            torch.stack(
                [
                    get_variable(x, ti)
                    for ti in ["t500", "t400", "t300", "t250", "t200"]
                ],
                dim=1,
            ),
            dim=1,
        )
        # Get z200 - z850 width
        dz_200_850 = get_variable(x, "z200") - get_variable(x, "z850")

        if self.lat_threshold is not None:
            w10m = w10m[:, indices].reshape(x.shape[0], -1, nlon)
            msl = msl[:, indices].reshape(x.shape[0], -1, nlon)
            vort850 = vort850[:, indices].reshape(x.shape[0], -1, nlon)
            t_200_500_mean = t_200_500_mean[:, indices].reshape(x.shape[0], -1, nlon)
            dz_200_850 = dz_200_850[:, indices].reshape(x.shape[0], -1, nlon)

        outs = []
        for i in range(x.shape[0]):
            centers = self._find_centers(
                lat,
                lon,
                msl[i],
                w10m[i],
                vort850[i],
                t_200_500_mean[i],
                dz_200_850[i],
                vorticity_threshold=self.vorticity_threshold,
                mslp_threshold=self.msl_threshold,
                temp_dec_threshold=self.temp_dec_threshold,
                exclude_border=self.exclude_border,
            )
            outs.append(centers)

        # https://pytorch.org/docs/stable/generated/torch.nn.utils.rnn.pad_sequence.html
        out = torch.nn.utils.rnn.pad_sequence(
            outs, padding_value=self.PATH_FILL_VALUE, batch_first=True
        )
        # [batch, path_id, step, variable]
        self.path_buffer = self.append_paths(
            out,
            self.path_buffer,
            self.path_search_distance,
            self.path_search_window_size,
        )
        out = torch.where(
            self.path_buffer == self.PATH_FILL_VALUE, torch.nan, self.path_buffer
        )

        output_coords = self.output_coords(coords)
        output_coords["path_id"] = np.arange(self.path_buffer.shape[1])
        output_coords["step"] = np.arange(self.path_buffer.shape[2])

        return out, output_coords



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Protocol, runtime_checkable

import torch

from earth2studio.utils.type import CoordSystem


@runtime_checkable
class IOBackend(Protocol):
    """Interface for a generic IO backend."""

    def __init__(
        self,
    ) -> None:
        pass

    def add_array(
        self, coords: CoordSystem, array_name: str | list[str], **kwargs: dict[str, Any]
    ) -> None:
        """
        Add an array with `array_name` to the existing IO backend object.

        Parameters
        ----------
        coords : OrderedDict
            Ordered dictionary of representing the dimensions and coordinate data
            of x.
        array_name : str
            Name of the arrays that will be initialized with coordinates as dimensions.
        kwargs : dict[str, Any], optional
            Optional keyword arguments that will be passed to the IO backend constructor.
        """
        pass

    def write(
        self,
        x: torch.Tensor | list[torch.Tensor],
        coords: CoordSystem,
        array_name: str | list[str],
    ) -> None:
        """
        Write data to the current backend using the passed array_name.

        Parameters
        ----------
        x : torch.Tensor | list[torch.Tensor]
            Tensor(s) to be written to zarr store.
        coords : OrderedDict
            Coordinates of the passed data.
        array_name : str | list[str]
            Name(s) of the array(s) that will be written to.
        """
        pass



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Source: https://github.com/andregraubner/ClimateNet
# MIT License
#
# Copyright (c) 2020 André Graubner
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


###########################################################################
# CGNet: A Light-weight Context Guided Network for Semantic Segmentation
# Paper-Link: https://arxiv.org/pdf/1811.08201.pdf
###########################################################################

import torch
import torch.nn as nn
import torch.nn.functional as F


class Wrap(torch.nn.Module):
    """Climate net wrapper for padding."""

    def __init__(self, padding):
        super().__init__()
        self.p = padding

    def forward(self, x):
        # creating the circular padding
        return F.pad(x, (self.p,) * 4, mode="circular")


class ConvBNPReLU(nn.Module):
    """Convolutional Net with Batch Norm and PreLU."""

    def __init__(self, nIn, nOut, kSize, stride=1):
        """
        args:
            nIn: number of input channels
            nOut: number of output channels
            kSize: kernel size
            stride: stride rate for down-sampling. Default is 1
        """
        super().__init__()
        padding = int((kSize - 1) / 2)
        self.padding = Wrap(padding=padding)
        self.conv = nn.Conv2d(nIn, nOut, (kSize, kSize), stride=stride, bias=False)
        self.bn = nn.BatchNorm2d(nOut, eps=1e-03)
        self.act = nn.PReLU(nOut)

    def forward(self, input):
        """
        args:
           input: input feature map
           return: transformed feature map
        """
        output = self.padding(input)
        output = self.conv(output)
        output = self.bn(output)
        output = self.act(output)
        return output


class BNPReLU(nn.Module):
    """Batch Norm with PReLU layer."""

    def __init__(self, nOut):
        """
        args:
           nOut: channels of output feature maps
        """
        super().__init__()
        self.bn = nn.BatchNorm2d(nOut, eps=1e-03)
        self.act = nn.PReLU(nOut)

    def forward(self, input):
        """
        args:
           input: input feature map
           return: normalized and thresholded feature map
        """
        output = self.bn(input)
        output = self.act(output)
        return output


class ConvBN(nn.Module):
    """Convolutional Layer with Batch Norm."""

    def __init__(self, nIn, nOut, kSize, stride=1):
        """
        args:
           nIn: number of input channels
           nOut: number of output channels
           kSize: kernel size
           stride: optinal stide for down-sampling
        """
        super().__init__()
        padding = int((kSize - 1) / 2)
        self.padding = Wrap(padding=padding)
        self.conv = nn.Conv2d(nIn, nOut, (kSize, kSize), stride=stride, bias=False)
        self.bn = nn.BatchNorm2d(nOut, eps=1e-03)

    def forward(self, input):
        """
        args:
           input: input feature map
           return: transformed feature map
        """
        output = self.padding(input)
        output = self.conv(output)
        output = self.bn(output)
        return output


class Conv(nn.Module):
    """Ordinary Convolutional Layer."""

    def __init__(self, nIn, nOut, kSize, stride=1):
        """
        args:
            nIn: number of input channels
            nOut: number of output channels
            kSize: kernel size
            stride: optional stride rate for down-sampling
        """
        super().__init__()
        padding = int((kSize - 1) / 2)
        self.padding = Wrap(padding=padding)
        self.conv = nn.Conv2d(nIn, nOut, (kSize, kSize), stride=stride, bias=False)

    def forward(self, input):
        """
        args:
           input: input feature map
           return: transformed feature map
        """
        output = self.padding(input)
        output = self.conv(output)
        return output


class ChannelWiseConv(nn.Module):
    """Channel Wise convolutional layer."""

    def __init__(self, nIn, nOut, kSize, stride=1):
        """
        Args:
            nIn: number of input channels
            nOut: number of output channels, default (nIn == nOut)
            kSize: kernel size
            stride: optional stride rate for down-sampling
        """
        super().__init__()
        padding = int((kSize - 1) / 2)
        self.padding = Wrap(padding=padding)
        self.conv = nn.Conv2d(
            nIn, nOut, (kSize, kSize), stride=stride, groups=nIn, bias=False
        )

    def forward(self, input):
        """
        args:
           input: input feature map
           return: transformed feature map
        """
        output = self.padding(input)
        output = self.conv(output)
        return output


class DilatedConv(nn.Module):
    """Convolutional Layer with Dilation."""

    def __init__(self, nIn, nOut, kSize, stride=1, d=1):
        """
        args:
           nIn: number of input channels
           nOut: number of output channels
           kSize: kernel size
           stride: optional stride rate for down-sampling
           d: dilation rate
        """
        super().__init__()
        padding = int((kSize - 1) / 2) * d
        self.padding = Wrap(padding=padding)
        self.conv = nn.Conv2d(
            nIn, nOut, (kSize, kSize), stride=stride, bias=False, dilation=d
        )

    def forward(self, input):
        """
        args:
           input: input feature map
           return: transformed feature map
        """
        output = self.padding(input)
        output = self.conv(output)
        return output


class ChannelWiseDilatedConv(nn.Module):
    """Channel-wise Convolutional Layer with Dilation."""

    def __init__(self, nIn, nOut, kSize, stride=1, d=1):
        """
        args:
           nIn: number of input channels
           nOut: number of output channels, default (nIn == nOut)
           kSize: kernel size
           stride: optional stride rate for down-sampling
           d: dilation rate
        """
        super().__init__()
        padding = int((kSize - 1) / 2) * d
        self.padding = Wrap(padding=padding)
        self.conv = nn.Conv2d(
            nIn, nOut, (kSize, kSize), stride=stride, groups=nIn, bias=False, dilation=d
        )

    def forward(self, input):
        """
        args:
           input: input feature map
           return: transformed feature map
        """
        output = self.padding(input)
        output = self.conv(output)
        return output


class FGlo(nn.Module):
    """The FGlo class is employed to refine the joint feature of both local feature and
    surrounding context."""

    def __init__(self, channel, reduction=16):
        super().__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Sequential(
            nn.Linear(channel, channel // reduction),
            nn.ReLU(inplace=True),
            nn.Linear(channel // reduction, channel),
            nn.Sigmoid(),
        )

    def forward(self, x):
        b, c, _, _ = x.size()
        y = self.avg_pool(x).view(b, c)
        y = self.fc(y).view(b, c, 1, 1)
        return x * y


class ContextGuidedBlock_Down(nn.Module):
    """The size of feature map divided 2, (H,W,C)---->(H/2, W/2, 2C)"""

    def __init__(self, nIn, nOut, dilation_rate=2, reduction=16):
        """
        args:
           nIn: the channel of input feature map
           nOut: the channel of output feature map, and nOut=2*nIn
        """
        super().__init__()
        self.conv1x1 = ConvBNPReLU(nIn, nOut, 3, 2)  # size/2, channel: nIn--->nOut

        self.F_loc = ChannelWiseConv(nOut, nOut, 3, 1)
        self.F_sur = ChannelWiseDilatedConv(nOut, nOut, 3, 1, dilation_rate)

        self.bn = nn.BatchNorm2d(2 * nOut, eps=1e-3)
        self.act = nn.PReLU(2 * nOut)
        self.reduce = Conv(2 * nOut, nOut, 1, 1)  # reduce dimension: 2*nOut--->nOut

        self.F_glo = FGlo(nOut, reduction)

    def forward(self, input):
        output = self.conv1x1(input)
        loc = self.F_loc(output)
        sur = self.F_sur(output)

        joi_feat = torch.cat([loc, sur], 1)  # the joint feature
        joi_feat = self.bn(joi_feat)
        joi_feat = self.act(joi_feat)
        joi_feat = self.reduce(joi_feat)  # channel= nOut

        output = self.F_glo(joi_feat)  # F_glo is employed to refine the joint feature

        return output


class ContextGuidedBlock(nn.Module):
    """Context Guided Block."""

    def __init__(self, nIn, nOut, dilation_rate=2, reduction=16, add=True):
        """
        args:
           nIn: number of input channels
           nOut: number of output channels,
           add: if true, residual learning
        """
        super().__init__()
        n = int(nOut / 2)
        self.conv1x1 = ConvBNPReLU(
            nIn, n, 1, 1
        )  # 1x1 Conv is employed to reduce the computation
        self.F_loc = ChannelWiseConv(n, n, 3, 1)  # local feature
        self.F_sur = ChannelWiseDilatedConv(
            n, n, 3, 1, dilation_rate
        )  # surrounding context
        self.bn_prelu = BNPReLU(nOut)
        self.add = add
        self.F_glo = FGlo(nOut, reduction)

    def forward(self, input):
        output = self.conv1x1(input)
        loc = self.F_loc(output)
        sur = self.F_sur(output)

        joi_feat = torch.cat([loc, sur], 1)

        joi_feat = self.bn_prelu(joi_feat)

        output = self.F_glo(joi_feat)  # F_glo is employed to refine the joint feature
        # if residual version
        if self.add:
            output = input + output
        return output


class InputInjection(nn.Module):
    """Inject Input with pooling."""

    def __init__(self, downsamplingRatio):
        super().__init__()
        self.pool = nn.ModuleList()
        for i in range(0, downsamplingRatio):
            self.pool.append(Wrap(padding=1))
            self.pool.append(nn.AvgPool2d(3, stride=2))

    def forward(self, input):
        for pool in self.pool:
            input = pool(input)
        return input


class CGNetModule(nn.Module):
    """
    CGNet (Wu et al, 2018: https://arxiv.org/pdf/1811.08201.pdf) implementation.
    This is taken from their implementation, we do not claim credit for this.
    """

    def __init__(self, classes=19, channels=4, M=3, N=21, dropout_flag=False):
        """
        args:
          classes: number of classes in the dataset. Default is 19 for the cityscapes
          M: the number of blocks in stage 2
          N: the number of blocks in stage 3
        """
        super().__init__()
        self.level1_0 = ConvBNPReLU(
            channels, 32, 3, 2
        )  # feature map size divided 2, 1/2
        self.level1_1 = ConvBNPReLU(32, 32, 3, 1)
        self.level1_2 = ConvBNPReLU(32, 32, 3, 1)

        self.sample1 = InputInjection(1)  # down-sample for Input Injection, factor=2
        self.sample2 = InputInjection(2)  # down-sample for Input Injiection, factor=4

        self.b1 = BNPReLU(32 + channels)

        # stage 2
        self.level2_0 = ContextGuidedBlock_Down(
            32 + channels, 64, dilation_rate=2, reduction=8
        )
        self.level2 = nn.ModuleList()
        for i in range(0, M - 1):
            self.level2.append(
                ContextGuidedBlock(64, 64, dilation_rate=2, reduction=8)
            )  # CG block
        self.bn_prelu_2 = BNPReLU(128 + channels)

        # stage 3
        self.level3_0 = ContextGuidedBlock_Down(
            128 + channels, 128, dilation_rate=4, reduction=16
        )
        self.level3 = nn.ModuleList()
        for i in range(0, N - 1):
            self.level3.append(
                ContextGuidedBlock(128, 128, dilation_rate=4, reduction=16)
            )  # CG block
        self.bn_prelu_3 = BNPReLU(256)

        if dropout_flag:
            self.classifier = nn.Sequential(
                nn.Dropout2d(0.1, False), Conv(256, classes, 1, 1)
            )
        else:
            self.classifier = nn.Sequential(Conv(256, classes, 1, 1))

        # init weights
        for m in self.modules():
            classname = m.__class__.__name__
            if classname.find("Conv2d") != -1:
                nn.init.kaiming_normal_(m.weight)
                if m.bias is not None:
                    m.bias.data.zero_()
                elif classname.find("ConvTranspose2d") != -1:
                    nn.init.kaiming_normal_(m.weight)
                    if m.bias is not None:
                        m.bias.data.zero_()

    def forward(self, input):
        """
        args:
            input: Receives the input RGB image
            return: segmentation map
        """
        # stage 1
        output0 = self.level1_0(input)
        output0 = self.level1_1(output0)
        output0 = self.level1_2(output0)
        inp1 = self.sample1(input)
        inp2 = self.sample2(input)

        # stage 2
        output0_cat = self.b1(torch.cat([output0, inp1], 1))
        output1_0 = self.level2_0(output0_cat)  # down-sampled

        for i, layer in enumerate(self.level2):
            if i == 0:
                output1 = layer(output1_0)
            else:
                output1 = layer(output1)

        output1_cat = self.bn_prelu_2(torch.cat([output1, output1_0, inp2], 1))

        # stage 3
        output2_0 = self.level3_0(output1_cat)  # down-sampled
        for i, layer in enumerate(self.level3):
            if i == 0:
                output2 = layer(output2_0)
            else:
                output2 = layer(output2)

        output2_cat = self.bn_prelu_3(torch.cat([output2_0, output2], 1))

        # classifier
        classifier = self.classifier(output2_cat)

        # upsample segmenation map ---> the input image size
        out = F.interpolate(
            classifier, input.size()[2:], mode="bilinear", align_corners=False
        )  # Upsample score map, factor=8
        return out



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable


class LexiconType(type):
    """Lexicon."""

    def __getitem__(cls, val: str) -> tuple[str, Callable]:
        """Retrieve variable name."""
        return cls.get_item(val)  # type: ignore[attr-defined]


E2STUDIO_VOCAB = {
    "u10m": "u-component (eastward, zonal) of wind at 10 m",
    "v10m": "v-component (northward, meridional) of wind at 10 m",
    "u100m": "u-component of wind at 100 m",
    "v100m": "v-component of wind at 100 m",
    "t2m": "temperature at 2m",
    "d2m": "dewpoint temperature at 2m",
    "r2m": "relative humidity at 2 m",
    "q2m": "specific humidity at 2 m",
    "sp": "surface pressure",
    "msl": "mean sea level pressure",
    "tcwv": "total column water vapor / precipitable water (kg m-2)",
    "tcw": "total column water vapor (kg m-2)",
    "tp": "total precipitation (m)",
    "tpp": "total precipitation probability",
    "tpi": "total precipitation index",
    "tp06": "total precipitation accumulated over past 6 hours",
    "fg10m": "maximum 10 m wind gust since previous post-processing",
    "lsm": "land-sea mask",
    "zsl": "geopotential at mean sea-level",
    "uvcossza": "cosine of solar zenith angle",
    "refc": "Maximum/Composite radar reflectivity (dB)",
    "csnow": "categorical snow",
    "cicep": "categorical (80, 90, or 100  percent chance of measurable precipitation (0.01 inch)) ice pellets",
    "cfrzr": "categorical (80, 90, or 100  percent chance of measurable precipitation (0.01 inch)) freezing rain",
    "crain": "categorical (80, 90, or 100  percent chance of measurable precipitation (0.01 inch)) rain",
    "tau300-700": "Geopotential thickness between 300 and 700 hPa levels (z300 - z700)",
    "tc_lat": "tropical cyclone center latitude position",
    "tc_lon": "tropical cyclone center longitude position",
    "tc_msl": "tropical cyclone center min mean sea level pressure",
    "tc_w10m": "tropical cyclone center max wind speed at 10m",
    "u50": "u-component of wind at 50 hPa",
    "u100": "u-component of wind at 100 hPa",
    "u150": "u-component of wind at 150 hPa",
    "u200": "u-component of wind at 200 hPa",
    "u250": "u-component of wind at 250 hPa",
    "u300": "u-component of wind at 300 hPa",
    "u400": "u-component of wind at 400 hPa",
    "u500": "u-component of wind at 500 hPa",
    "u600": "u-component of wind at 600 hPa",
    "u700": "u-component of wind at 700 hPa",
    "u850": "u-component of wind at 850 hPa",
    "u925": "u-component of wind at 925 hPa",
    "u1000": "u-component of wind at 1000 hPa",
    "v50": "v-component of wind at 50 hPa",
    "v100": "v-component of wind at 100 hPa",
    "v150": "v-component of wind at 150 hPa",
    "v200": "v-component of wind at 200 hPa",
    "v250": "v-component of wind at 250 hPa",
    "v300": "v-component of wind at 300 hPa",
    "v400": "v-component of wind at 400 hPa",
    "v500": "v-component of wind at 500 hPa",
    "v600": "v-component of wind at 600 hPa",
    "v700": "v-component of wind at 700 hPa",
    "v850": "v-component of wind at 850 hPa",
    "v925": "v-component of wind at 925 hPa",
    "v1000": "v-component of wind at 1000 hPa",
    "w50": "vertical wind at 50 hPa",
    "w100": "vertical wind at 100 hPa",
    "w150": "vertical wind at 150 hPa",
    "w200": "vertical wind at 200 hPa",
    "w250": "vertical wind at 250 hPa",
    "w300": "vertical wind at 300 hPa",
    "w400": "vertical wind at 400 hPa",
    "w500": "vertical wind at 500 hPa",
    "w600": "vertical wind at 600 hPa",
    "w700": "vertical wind at 700 hPa",
    "w850": "vertical wind at 850 hPa",
    "w925": "vertical wind at 925 hPa",
    "w1000": "vertical wind at 1000 hPa",
    "z50": "geopotential at 50 hPa",
    "z100": "geopotential at 100 hPa",
    "z150": "geopotential at 150 hPa",
    "z200": "geopotential at 200 hPa",
    "z250": "geopotential at 250 hPa",
    "z300": "geopotential at 300 hPa",
    "z400": "geopotential at 400 hPa",
    "z500": "geopotential at 500 hPa",
    "z600": "geopotential at 600 hPa",
    "z700": "geopotential at 700 hPa",
    "z850": "geopotential at 850 hPa",
    "z925": "geopotential at 925 hPa",
    "z1000": "geopotential at 1000 hPa",
    "t50": "temperature at 50 hPa",
    "t100": "temperature at 100 hPa",
    "t150": "temperature at 150 hPa",
    "t200": "temperature at 200 hPa",
    "t250": "temperature at 250 hPa",
    "t300": "temperature at 300 hPa",
    "t400": "temperature at 400 hPa",
    "t500": "temperature at 500 hPa",
    "t600": "temperature at 600 hPa",
    "t700": "temperature at 700 hPa",
    "t850": "temperature at 850 hPa",
    "t925": "temperature at 925 hPa",
    "t1000": "temperature at 1000 hPa",
    "r50": "relative humidity at 50 hPa",
    "r100": "relative humidity at 100 hPa",
    "r150": "relative humidity at 150 hPa",
    "r200": "relative humidity at 200 hPa",
    "r250": "relative humidity at 250 hPa",
    "r300": "relative humidity at 300 hPa",
    "r400": "relative humidity at 400 hPa",
    "r500": "relative humidity at 500 hPa",
    "r600": "relative humidity at 600 hPa",
    "r700": "relative humidity at 700 hPa",
    "r850": "relative humidity at 850 hPa",
    "r925": "relative humidity at 925 hPa",
    "r1000": "relative humidity at 1000 hPa",
    "q50": "specific humidity at 50 hPa",
    "q100": "specific humidity at 100 hPa",
    "q150": "specific humidity at 150 hPa",
    "q200": "specific humidity at 200 hPa",
    "q250": "specific humidity at 250 hPa",
    "q300": "specific humidity at 300 hPa",
    "q400": "specific humidity at 400 hPa",
    "q500": "specific humidity at 500 hPa",
    "q600": "specific humidity at 600 hPa",
    "q700": "specific humidity at 700 hPa",
    "q850": "specific humidity at 850 hPa",
    "q925": "specific humidity at 925 hPa",
    "q1000": "specific humidity at 1000 hPa",
    "vpd100": "vapor pressure deficit (hPa) at 100 hPa",
    "vpd200": "vapor pressure deficit (hPa) at 200 hPa",
    "vpd300": "vapor pressure deficit (hPa) at 300 hPa",
    "vpd400": "vapor pressure deficit (hPa) at 400 hPa",
    "vpd500": "vapor pressure deficit (hPa) at 500 hPa",
    "vpd600": "vapor pressure deficit (hPa) at 600 hPa",
    "vpd700": "vapor pressure deficit (hPa) at 700 hPa",
    "vpd850": "vapor pressure deficit (hPa) at 850 hPa",
    "vpd925": "vapor pressure deficit (hPa) at 925 hPa",
    "vpd1000": "vapor pressure deficit (hPa) at 1000 hPa",
    "ws100": "wind speed at 100 hPa",
    "ws200": "wind speed at 200 hPa",
    "ws300": "wind speed at 300 hPa",
    "ws400": "wind speed at 400 hPa",
    "ws500": "wind speed at 500 hPa",
    "ws600": "wind speed at 600 hPa",
    "ws700": "wind speed at 700 hPa",
    "ws850": "wind speed at 850 hPa",
    "ws925": "wind speed at 925 hPa",
    "ws1000": "wind speed at 1000 hPa",
    "skt": "skin temperature at surface",
    "slor": "slope of sub-grid scale orography",
    "sdor": "standard deviation of sub-grid scale orography",
    "swvl1": "volumetric soil water layer 1 (m³/m³)",  # Ask SAs for name confirmation, altertative (vsw?)
    "swvl2": "volumetric soil water layer 2 (m³/m³)",  # Ask SAs for name confirmation, altertative (vsw?)
    "stl1": "soil temperature level 1 (K)",  # Ask SAs for name confirmation, altertative (sot?!?!)
    "stl2": "soil temperature level 2 (K)",  # Ask SAs for name confirmation, altertative (sot?!?!)
}



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import OrderedDict
from datetime import datetime, timezone

import numpy as np
import torch
import xarray as xr

from earth2studio.utils.imports import check_extra_imports

try:
    from physicsnemo import Module
    from physicsnemo.utils.zenith_angle import cos_zenith_angle
except ImportError:
    Module = None
    cos_zenith_angle = None

from earth2studio.models.auto import AutoModelMixin, Package
from earth2studio.models.batch import batch_coords, batch_func
from earth2studio.models.dx.base import DiagnosticModel
from earth2studio.utils import (
    handshake_coords,
    handshake_dim,
)
from earth2studio.utils.type import CoordSystem

VARIABLES = [
    "u100m",
    "v100m",
    "msl",
    "u300",
    "u850",
    "u925",
    "v300",
    "v850",
    "v925",
    "z300",
    "z850",
    "z925",
    "t850",
    "t925",
    "q500",
    "q850",
    "q925",
]


@check_extra_imports("windgust-afno", [Module, cos_zenith_angle])
class WindgustAFNO(torch.nn.Module, AutoModelMixin):
    """Wind gust AFNO diagnsotic model. Predicts the maximum wind gust during the
    preceding hour with the units m/s. This model uses an 17 atmospheric inputs and
    outputs one on a 0.25 degree lat-lon grid (south-pole excluding) [720 x 1440].

    Note
    ----
    For more information on the model, please refer to:

    - https://catalog.ngc.nvidia.com/orgs/nvidia/teams/earth-2/models/afno_dx_wg-v1-era5

    Parameters
    ----------
    core_model : torch.nn.Module
        Core pytorch model
    landsea_mask : torch.Tensor
        Land sea mask tensor of size [720,1440]
    orography : torch.Tensor
        Surface geopotential (orography) tensor of size [720,1440]
    center : torch.Tensor
        Model center normalization tensor of size [17,1,1]
    scale : torch.Tensor
        Model scale normalization tensor of size [17,1,1]
    """

    def __init__(
        self,
        core_model: torch.nn.Module,
        landsea_mask: torch.Tensor,
        orography: torch.Tensor,
        center: torch.Tensor,
        scale: torch.Tensor,
    ):
        super().__init__()
        self.core_model = core_model
        self.register_buffer("center", center)
        self.register_buffer("scale", scale)
        self.register_buffer("landsea_mask", landsea_mask)
        self.register_buffer(
            "topographic_height",
            (orography),
        )

    def input_coords(self) -> CoordSystem:
        """Input coordinate system of diagnostic model

        Returns
        -------
        CoordSystem
            Coordinate system dictionary
        """
        return OrderedDict(
            {
                "batch": np.empty(0),
                "time": np.empty(0),
                "lead_time": np.empty(0),
                "variable": np.array(VARIABLES),
                "lat": np.linspace(90, -90, 720, endpoint=False),
                "lon": np.linspace(0, 360, 1440, endpoint=False),
            }
        )

    @batch_coords()
    def output_coords(self, input_coords: CoordSystem) -> CoordSystem:
        """Output coordinate system of diagnostic model

        Parameters
        ----------
        input_coords : CoordSystem
            Input coordinate system to transform into output_coords
            by default None, will use self.input_coords.

        Returns
        -------
        CoordSystem
            Coordinate system dictionary
        """
        target_input_coords = self.input_coords()
        handshake_dim(input_coords, "lon", 5)
        handshake_dim(input_coords, "lat", 4)
        handshake_dim(input_coords, "variable", 3)
        handshake_coords(input_coords, target_input_coords, "lon")
        handshake_coords(input_coords, target_input_coords, "lat")
        handshake_coords(input_coords, target_input_coords, "variable")

        output_coords = input_coords.copy()
        output_coords["variable"] = np.array(["fg10m"])
        return output_coords

    def __str__(self) -> str:
        return "windgustnet"

    @classmethod
    def load_default_package(cls) -> Package:
        """Default pre-trained climatenet model package from Nvidia model registry"""
        return Package(
            "ngc://models/nvidia/earth-2/afno_dx_wg-v1-era5@v0.1.0",
            cache_options={
                "cache_storage": Package.default_cache("climatenet"),
                "same_names": True,
            },
        )

    @classmethod
    def load_model(cls, package: Package) -> DiagnosticModel:
        """Load diagnostic from package"""
        model = Module.from_checkpoint(package.resolve("afno_windgust_1h.mdlus"))
        model.eval()

        input_center = torch.Tensor(np.load(package.resolve("global_means.npy")))

        input_scale = torch.Tensor(np.load(package.resolve("global_stds.npy")))
        lsm = torch.Tensor(
            xr.open_dataset(package.resolve("land_sea_mask.nc"))["LSM"].values
        )[None, :, :-1]

        orography = torch.Tensor(
            xr.open_dataset(package.resolve("orography.nc"))["Z"].values
        )[None, :, :-1]
        orography = (orography - orography.mean()) / orography.std()

        return cls(model, lsm, orography, input_center, input_scale)

    def _compute_sza(
        self,
        lon: np.array,
        lat: np.array,
        time: np.datetime64,
        lead_time: np.timedelta64,
    ) -> torch.Tensor:
        """Compute solar zenith angle"""
        _unix = np.datetime64(0, "s")
        _ds = np.timedelta64(1, "s")
        t = time + lead_time
        t = datetime.fromtimestamp((t - _unix) / _ds, tz=timezone.utc)
        return torch.Tensor(cos_zenith_angle(t, lon, lat))

    @torch.inference_mode()
    @batch_func()
    def __call__(
        self,
        x: torch.Tensor,
        coords: CoordSystem,
    ) -> tuple[torch.Tensor, CoordSystem]:
        """Forward pass of diagnostic"""
        output_coords = self.output_coords(coords)
        out = torch.zeros_like(x[..., :1, :, :])
        x = (x - self.center) / self.scale

        grid_x, grid_y = torch.meshgrid(
            torch.tensor(coords["lat"]), torch.tensor(coords["lon"])
        )

        # compute solar zenith angle and concatenate
        for j, _ in enumerate(coords["batch"]):
            for k, t in enumerate(coords["time"]):
                for lt, dt in enumerate(coords["lead_time"]):
                    sza = (
                        self._compute_sza(grid_x, grid_y, t, dt)
                        .unsqueeze(0)
                        .unsqueeze(0)
                        .to(x.device)
                    )
                    tran = torch.cat(
                        [sza, self.topographic_height, self.landsea_mask], dim=1
                    )
                    in_ = torch.cat((x[j, k, lt : lt + 1], tran), dim=1)
                    out[j, k, lt : lt + 1] = self.core_model(in_)

        out = torch.clamp(out, min=0)
        return out, output_coords



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable

import numpy as np

from .base import LexiconType


class IMERGLexicon(metaclass=LexiconType):
    """IMERG Lexicon"""

    VOCAB = {
        "tp": "precipitation",
        "tpp": "probabilityLiquidPrecipitation",
        "tpi": "precipitationQualityIndex",
    }

    @classmethod
    def get_item(cls, val: str) -> tuple[str, Callable]:
        """Return name in IMERG vocabulary."""
        imerg_key = cls.VOCAB[val]
        if val == "tp":
            # IMERG is mm/hr by default, convert to meters to match ECMWF
            # https://arthurhou.pps.eosdis.nasa.gov/Documents/IMERG_TechnicalDocumentation_final.pdf
            def mod(x: np.array) -> np.array:
                """Modify data value (if necessary)."""
                return x / 1000.0

        else:

            def mod(x: np.array) -> np.array:
                """Modify name (if necessary)."""
                return x

        return imerg_key, mod
